type op = Add | Sub | Equals | NotEquals | Lt | Lteq | Gt | Gteq | And | Or | Amp | Eq | Neq


type expr = 
	| IntLiteral of int
	| StringLiteral of string
	| Id of string
	| ArrayElement of string * int
	| ArrayElementViaId of string * string
	| Binop of expr * op * expr
	| Assign of string * expr
	| AssignArrayElement of string * int * expr	
	| AppendFile of expr * expr
	| GetLine of expr * expr 
	| Print of expr
	| RegexMatch of expr * expr
	| RegexReplace of expr * expr * expr
	| StrReplace of expr * expr * expr
	| Call of string


type stmt =
	| Block of stmt list
	| Expr of expr
	| If of expr * stmt list * stmt list
	| While of expr * stmt list


type var_type = Int | IntArray | String | StringArray


type var_decl = {
	vname: string; (* variable name *)
	vtype: var_type; (* variable type *)
	vsize: int; (* variable size *)	
	}


type func_decl = {
	fname: string;  (* function name *)
	fbody: stmt list (* list of stmts in function body *)
	}

		
type program = var_decl list * func_decl list (* global vars, funcs  *)	