open Ast

(* ================================================== *)
module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

(* ================================================== *)
exception ReturnException of string * (int array NameMap.t * string array NameMap.t * func_decl NameMap.t)


let string_of_char = String.make 1;;

(* ============================================================= *)
(* Evaluate an expression and return (value, updated environment) *)
let rec eval env = function
	| IntLiteral(i) -> string_of_int(i), env
	| StringLiteral(s) -> s, env
	(* -------------------------------------------------- *)  
  | Id(var) ->			
			let (int_var_decls, string_var_decls, func_decls) = env 
			in
			if NameMap.mem var int_var_decls then				
				string_of_int((NameMap.find var int_var_decls).(0)), env				
			else if NameMap.mem var string_var_decls then
			  ((NameMap.find var string_var_decls).(0)), env
			else if NameMap.mem var func_decls then
				let fdecl = (NameMap.find var func_decls)
				in
				"0", call fdecl env
			else
				raise (Failure ("undeclared identifier " ^ var))
	(* -------------------------------------------------- *)
	| ArrayElement(s, i) ->
			let (int_var_decls, string_var_decls, func_decls) = env 
			in
			if NameMap.mem s int_var_decls then
				string_of_int((NameMap.find s int_var_decls).(i)), env				
			else if NameMap.mem s string_var_decls then
			  ((NameMap.find s string_var_decls).(i)), env				
			else			  
		    raise (Failure ("undeclared identifier " ^ s))
	(* -------------------------------------------------- *)
	| ArrayElementViaId(s1, s2) ->
			let (int_var_decls, string_var_decls, func_decls) = env
			in
			let i =
				if NameMap.mem s2 int_var_decls then				
					(NameMap.find s2 int_var_decls).(0)				
				else
					raise (Failure ("undeclared identifier " ^ s2))
			in
			if NameMap.mem s1 int_var_decls then
				string_of_int((NameMap.find s1 int_var_decls).(i)), env				
			else if NameMap.mem s1 string_var_decls then
			  ((NameMap.find s1 string_var_decls).(i)), env				
			else			  
		    raise (Failure ("undeclared identifier " ^ s1))
	(* -------------------------------------------------- *)
  | Binop(e1, op, e2) ->
  		let v1, env = eval env e1 					
			in
    	let v2, env = eval env e2 
			in
  		let boolean i = if i then 0 else 1 
			in
  		(match op with
    	  |	Add -> string_of_int( int_of_string(v1) + int_of_string(v2) )
			  | Sub -> string_of_int( int_of_string(v1) - int_of_string(v2) )	  				  
			  | Equals -> string_of_int( boolean (int_of_string(v1) = int_of_string(v2)) )
			  | NotEquals -> string_of_int( boolean (int_of_string(v1) != int_of_string(v2)) )
			  | Lt -> string_of_int( boolean (int_of_string(v1) < int_of_string(v2)) )
			  | Lteq -> string_of_int( boolean (int_of_string(v1) <= int_of_string(v2)) )
			  | Gt -> string_of_int( boolean (int_of_string(v1) > int_of_string(v2)) )
			  | Gteq -> string_of_int( boolean (int_of_string(v1) >= int_of_string(v2)) )
				| And -> string_of_int( boolean( (int_of_string(v1) = 0) && (int_of_string(v2) = 0) ) )
				| Or -> string_of_int( boolean( (int_of_string(v1) = 0) || (int_of_string(v2) = 0) ) )
				| Amp -> v1 ^ v2
				| Eq -> string_of_int (boolean (v1 = v2))
				| Neq -> string_of_int (boolean (v1 <> v2))  
				), env
	(* -------------------------------------------------- *)
  | Assign(var, e) ->
				let v, (int_var_decls, string_var_decls, func_decls) = eval env e				
				in
				let assign_int (int_var_decls, string_var_decls, func_decls) var v =
					(NameMap.find var int_var_decls).(0) <- int_of_string(v);
					(int_var_decls, string_var_decls, func_decls)
				in
				let assign_string (int_var_decls, string_var_decls, func_decls) var v =
					(NameMap.find var string_var_decls).(0) <- v;
					(int_var_decls, string_var_decls, func_decls)
				in
				if NameMap.mem var int_var_decls then
					v, (assign_int (int_var_decls, string_var_decls, func_decls) var v)							
				else if NameMap.mem var string_var_decls then
					v, (assign_string (int_var_decls, string_var_decls, func_decls) var v)
				else
					raise (Failure ("undeclared identifier " ^ var))
	(* -------------------------------------------------- *)				
  | AssignArrayElement(var, i, e) ->
				let v, (int_var_decls, string_var_decls, func_decls) = eval env e				
				in
				let assign_int_arr_element (int_var_decls, string_var_decls, func_decls) var i v =
					(NameMap.find var int_var_decls).(i) <- int_of_string(v);
					(int_var_decls, string_var_decls, func_decls)
				in
				let assign_string_arr_element (int_var_decls, string_var_decls, func_decls) var i v =
					(NameMap.find var string_var_decls).(i) <- v;
					(int_var_decls, string_var_decls, func_decls)
				in
				if NameMap.mem var int_var_decls then
					v, (assign_int_arr_element (int_var_decls, string_var_decls, func_decls) var i v)							
				else if NameMap.mem var string_var_decls then
					v, (assign_string_arr_element (int_var_decls, string_var_decls, func_decls) var i v)
				else
					raise (Failure ("undeclared identifier " ^ var))
	(* -------------------------------------------------- *)
	| AppendFile (e1, e2) ->								
				let filepath, env = eval env e1
				in
	    	let appendString, env = eval env e2 
				in
	  		let out_channel = open_out_gen [Open_append;Open_creat] 0o777 filepath 
				in
				output_string out_channel appendString;
				close_out out_channel;
				"0", env
  (* -------------------------------------------------- *)
	| GetLine (e1, e2) ->			  
				let filepath, env = eval env e1
				in
	    	let lineNumber, env = eval env e2
				in				
				(* ---------------------- *)
				(* Define helper function *)
				let rec input_line_n in_channel curr_line_num targ_line_num =					
				  let line = input_line in_channel 
					in			  				  
				  if curr_line_num >= targ_line_num then
				    line, env
				  else
				    input_line_n in_channel (curr_line_num+1) targ_line_num;					
					in
				(* ---------------------- *)									
				  (try
						let in_channel = open_in_gen [Open_rdonly] 0o777 filepath
						in
						try							
						  let v, env = input_line_n in_channel 1 (int_of_string lineNumber);
							in
							close_in in_channel; 
							v, env
						with End_of_file -> close_in in_channel; ("", env)
					 with End_of_file -> ("", env))
								
	(* -------------------------------------------------- *)	
	| Print (e) ->
				let v, env = eval env e 
				in
		  	print_string (v);
				"0", env
				
	(* -------------------------------------------------- *)	
	| RegexMatch (e1, e2) ->
				let content, env = eval env e1
				in
	    	let regex_string, env = eval env e2				
				in
				let regex = Str.regexp regex_string 
				in
				(let rec srch_fwd_and_match regex content pos =
						try							
				  		let _ = Str.search_forward regex content pos
				  		in							
							let result = Str.matched_string content
							in							
							if result = "" then								 
							   srch_fwd_and_match regex content (pos+1)
								else
									result							
				 		with 
							| Not_found -> srch_fwd_and_match regex content (pos+1)
							| Invalid_argument(a) -> "" 						
				 in
				 srch_fwd_and_match regex content 0), env
	(* -------------------------------------------------- *)	
	| RegexReplace (e1, e2, e3) ->
				let content, env = eval env e1
				in
	    	let regex_string, env = eval env e2				
				in
				let regex = Str.regexp regex_string 
				in
				let rep_string, env = eval env e3 
				in
				(Str.global_replace regex rep_string content), env
	(* -------------------------------------------------- *)	
	| StrReplace (e1, e2, e3) ->
				let content, env = eval env e1
				in
	    	let match_string, env = eval env e2				
				in
				let regex = Str.regexp_string match_string 
				in
				let rep_string, env = eval env e3 
				in
				(Str.global_replace regex rep_string content), env									
	(* -------------------------------------------------- *)
  | Call(f) ->
		    let (int_var_decls, string_var_decls, func_decls) = env
				in
  			let fdecl =
    			try 
						NameMap.find f func_decls
    			with Not_found -> raise (Failure ("undefined function " ^ f))
  			in	  				  			
  			try
    			let env = call fdecl env 
					in 
					"0", env
  			with ReturnException(v, env) -> v, env

(* ===================================================== *)
(* Execute a statement and return an updated environment *)
and exec env = function
	| Block(stmts) -> List.fold_left exec env stmts
	(* -------------------------------------------------- *)
  | Expr(e) -> let _, env = eval env e 
	             in 
							 env
	(* -------------------------------------------------- *)
  | If(e, stmts1_rev, stmts2_rev) ->
		  let stmts1 = List.rev stmts1_rev
			in
			let stmts2 = List.rev stmts2_rev
			in
			(* First evaluate e and store the result in v *)
			let v, env = eval env e 
			in
			(* If v == 0, exec s1 otherwise exec s2 *)			
			List.fold_left exec env (if int_of_string(v) == 0 then 
															   stmts1 
															 else 
															   stmts2)
	(* -------------------------------------------------- *)
  | While(e, stmts_rev) ->
			let stmts = List.rev stmts_rev
			in
			let rec loop env =
				(* First evaluate e and store the result in v *)
    		let v, env = eval env e 
				in
				(* Keep looping if v == 0 then exec s *)
    		if int_of_string(v) == 0 then 
					loop (List.fold_left exec env stmts) 
				else 
					(* otherwise return updated env *)
					env
			in 					
			loop env

(* =========================================================== *)
(* Invoke a function and return an updated global symbol table *)
and call fdecl env =	      
	(* ====================================================================== *)          
	(* Execute each statement in sequence, return updated global symbol table *)
	List.fold_left exec env fdecl.fbody


(* =============================== *)
(* Main entry point: run a program *)
let run (vars, funcs) =	
	(* --------------------------------------------------- *)	
	(* Put int and intArray declarations in a symbol table *)
	(* Initialize ints and intArray elements to 0 *)
	let int_var_decls = List.fold_left
		                  (fun vars vdecl ->
												match vdecl.vtype with
													| Int | IntArray -> NameMap.add vdecl.vname (Array.make vdecl.vsize 0) vars													
													| _ -> vars
											)
		                  NameMap.empty vars
	in
	
	(* --------------------------------------------------------- *)	
	(* Put string and stringArray declarations in a symbol table *)
	(* Initialize strings and stringArray elements to "" *)
	let string_var_decls = List.fold_left
		                     (fun vars vdecl ->
												 		match vdecl.vtype with
													  | String | StringArray -> NameMap.add vdecl.vname (Array.make vdecl.vsize "") vars													
														| _ -> vars
												 )
		                  NameMap.empty vars
	in
	
	(* ------------------------------------------- *)
  (* Put function declarations in a symbol table *)
	let func_decls = List.fold_left
		               (fun funcs fdecl -> NameMap.add fdecl.fname fdecl funcs)
		               NameMap.empty funcs
  in	  	
		
	(* -------------------------------------------------------------------- *)
	(* Run a program: find and run "main" *)	
		
	try
	  call (NameMap.find "main" func_decls) (int_var_decls, string_var_decls, func_decls)
	with Not_found -> 
		raise (Failure ("did not find function 'main'"))
		



