%{ open Ast %}

%token PLUS MINUS
%token EQUALS NOTEQUALS LT LEQ GT GEQ AND OR ASSIGN AMP EQ NEQ
%token APPEND_FILE GET_LINE REGEX_MATCH PRINT REGEX_REP STR_REP
%token ELSE END FUNCTION IF THEN WHILE 
%token INT INTARRAY STRING STRINGARRAY  
%token EOF

%token <string> MAIN
%token <string> ID
%token <int> INDEX
%token <string> ID_INDEX
%token <int> INT_LITERAL
%token <string> STRING_LITERAL

%nonassoc ELSE
%nonassoc INDEX
%nonassoc ID_INDEX
%nonassoc NOELSE
%nonassoc APPEND_FILE GET_LINE PRINT REGEX_MATCH REGEX_REP STR_REP

%left ASSIGN
%left EQUALS NOTEQUALS EQ NEQ
%left LT LEQ GT GEQ AND OR
%left PLUS MINUS AMP

%start program
%type <Ast.program> program

%%

program:
	| /* nothing */ { ([], []) }
	| program vdecl { ($2 :: fst $1), snd $1 }
	| program fdecl { fst $1, ($2 :: snd $1) }  

vdecl:
	| INT ID                     { { vname = $2; vtype = Ast.Int; vsize = 1; } }
	| INTARRAY ID INT_LITERAL    { { vname = $2; vtype = Ast.IntArray; vsize = $3; } }
	| STRING ID                  { { vname = $2; vtype = Ast.String; vsize = 1; } }
	| STRINGARRAY ID INT_LITERAL { { vname = $2; vtype = Ast.StringArray; vsize = $3; } }

fdecl:	
	| FUNCTION ID stmt_list END   { {fname = $2; fbody = List.rev $3} }
	| FUNCTION MAIN stmt_list END { {fname = $2; fbody = List.rev $3} }

stmt_list:
	| /* nothing */  { [] }
	| stmt_list stmt { $2::$1 }

stmt:
	| expr                                      { Expr($1) }	
	| IF expr THEN stmt_list END %prec NOELSE   { If($2, $4, []) }
	| IF expr THEN stmt_list ELSE stmt_list END { If($2, $4, $6) }
	| WHILE expr THEN stmt_list END             { While($2, $4) } 
		
expr:
	| INT_LITERAL          { IntLiteral($1) }
	| STRING_LITERAL       { StringLiteral($1) }
	| ID                   { Id($1) }
	| ID INDEX             { ArrayElement($1, $2) }
	| ID ID_INDEX          { ArrayElementViaId($1, $2) }
  | ID ASSIGN expr       { Assign ($1, $3) }
	| ID INDEX ASSIGN expr { AssignArrayElement ($1, $2, $4) }

	| expr PLUS expr      { Binop($1, Add, $3) }
	| expr MINUS expr     { Binop($1, Sub, $3) }
	| expr EQUALS expr    { Binop($1, Equals, $3) }
	| expr NOTEQUALS expr { Binop($1, NotEquals, $3) }
	| expr LT expr        { Binop($1, Lt, $3) }
	| expr LEQ expr       { Binop($1, Lteq, $3) }
	| expr GT expr        { Binop($1, Gt, $3) }
	| expr GEQ expr       { Binop($1, Gteq, $3) }
	| expr AND expr       { Binop($1, And, $3) }
	| expr OR expr        { Binop($1, Or, $3) }
	| expr AMP expr       { Binop($1, Amp, $3) }
	| expr EQ expr        { Binop($1, Eq, $3) }
	| expr NEQ expr       { Binop($1, Neq, $3) }
	
	| APPEND_FILE expr expr    { AppendFile($2, $3) }
	| GET_LINE expr expr	     { GetLine($2, $3) }
	| PRINT expr               { Print($2) }
	| REGEX_MATCH expr expr    { RegexMatch($2, $3) }
	| REGEX_REP expr expr expr { RegexReplace($2, $3, $4) }
	| STR_REP expr expr expr   { StrReplace($2, $3, $4) }

