open Ast

let rec string_of_expr = function
	| StringLiteral(s) -> s
  | IntLiteral(l) -> string_of_int l
  | Id(s) -> s
	| ArrayElement(s, e) -> s ^ "[" ^ string_of_int(e) ^ "]"	
	| ArrayElementViaId(s1, s2) -> s1 ^ "[" ^ s2 ^ "]"
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
			| Add -> "+" 
			| Sub -> "-" 						
      | Equals -> "==" 
			| NotEquals -> "!="
      | Lt -> "<" 
			| Lteq -> "<=" 
			| Gt -> ">" 
			| Gteq -> ">="
			| And -> "and"
			| Or -> "or"
			| Amp -> "&"
			| Eq -> "eq"
			| Neq -> "neq"
			) ^ " " ^	
      string_of_expr e2	
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
	| AssignArrayElement(v, i, e) -> v ^ "[" ^ string_of_int i ^ "] = " ^ string_of_expr e	
	| AppendFile(e1, e2) -> "appendFile " ^ string_of_expr e1 ^ " " ^ string_of_expr e2
	| GetLine(e1, e2) -> "getLine " ^ string_of_expr e1 ^ " " ^ string_of_expr e2   
	| Print(e) -> "print " ^ string_of_expr e
	| RegexMatch(e1, e2) -> "regexMatch " ^ string_of_expr e1 ^ " " ^ string_of_expr e2
	| RegexReplace(e1, e2, e3) -> "regexReplace " ^ string_of_expr e1 ^ " " ^ string_of_expr e2 ^ " " ^ string_of_expr e3
	| StrReplace(e1, e2, e3) -> "strReplace " ^ string_of_expr e1 ^ " " ^ string_of_expr e2 ^ " " ^ string_of_expr e3
	| Call(f) -> f	

let rec string_of_stmt = function
    Block(stmts) -> "\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "\n"
  | Expr(expr) -> string_of_expr expr ^ "\n";  
  | If(e, stmts, []) -> "if " ^ string_of_expr e ^ " then \n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "\n end \n"
  | If(e, stmts1, stmts2) ->  "if " ^ string_of_expr e ^ " then \n" ^
                              "\n" ^ String.concat "" (List.map string_of_stmt stmts1) ^ "\n" ^ 
															"else\n" ^ 
															"\n" ^ String.concat "" (List.map string_of_stmt stmts2) ^ "end \n"  
  | While(e, stmts) -> "while " ^ string_of_expr e ^ " then \n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "\n end \n"

let string_of_var_type = function
	| Int -> "int "
	| String -> "string "
	| IntArray -> "intArray "
	| StringArray -> "stringArray "

let string_of_vdecl vdecl = 	
	(string_of_var_type vdecl.vtype) ^
	vdecl.vname ^ " \n"

let string_of_fdecl fdecl =
  fdecl.fname ^ "\n" ^ 
  String.concat "" (List.map string_of_stmt fdecl.fbody) ^ " \n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
