{ 
	open Parser;;
  let buffer = Buffer.create 16;;
	let char_for_decimal_code str = str.[0] <- '0'; Char.chr (int_of_string str land 0xFF);;
}

let alpha = ['a'-'z' 'A'-'Z']
let newline = '\n' | "\r\n"
let numeric = ['0'-'9']
let other = ['`' '~' '-' '_' '=' '+' 
             '[' '{' ']' '}' '|' 
						 ';' ':'
						 ',' '<' '.' '>' '/' '?']
let shift_numeric = ['!' '@' '#' '$' '%' '^' '&' '*' '(' ')']

let alphanumeric = alpha | numeric
let string = alpha | numeric | other | shift_numeric

rule token = parse

| '#'                             { comment lexbuf }

(* --------- *)
(* operators *)
| '+' 		            	      { PLUS }
| '-' 					              { MINUS }
| '[' (numeric+ as index) ']' { INDEX(int_of_string index) }
| '[' ((alpha+ alphanumeric*) as lxm) ']' { ID_INDEX(lxm) }
| "==" 					              { EQUALS }
| "!="                        { NEQ }
| '<'                         { LT }
| "<=" 	                      { LEQ }
| '>'                     	  { GT }
| ">=" 				                { GEQ }
| "and" 	               		  { AND }
| "or" 	                 		  { OR }
| '='                    			{ ASSIGN }
| '&'                    			{ AMP }
| "eq"                    		{ EQ }
| "neq"                    		{ NEQ }

| "main"                          { MAIN("main") }

(* ------------------ *)
(* built-in functions *)
| "appendFile"            				{ APPEND_FILE }
| "getLine"		              			{ GET_LINE }
| "print"                         { PRINT }
| "regexMatch"                    { REGEX_MATCH }
| "regexReplace"                  { REGEX_REP }
| "strReplace"                    { STR_REP }

(* ---------- *)
(* statements *)
| "else"                					{ ELSE }
| "end"                 					{ END }
| "function"	              			{ FUNCTION }
| "if"                            { IF }
| "then"                          { THEN }
| "while"                         { WHILE }

(* --------- *)
(* variables *)
| "int"                           { INT }
| "intArray"                      { INTARRAY }
| "string"                        { STRING }
| "stringArray"                   { STRINGARRAY }

(* -------- *)
(* literals *)
| numeric+ as lxm                     { INT_LITERAL(int_of_string lxm) }
| '"'                                 { Buffer.clear buffer; STRING_LITERAL(string_literal lexbuf) }
| (alpha+ alphanumeric*) as lxm       { ID(lxm) }

| eof                   					{ EOF }

| _                               { token lexbuf }

(* ------- *)
(* string literal *)
and string_literal = parse
| '"'                      { Buffer.contents buffer }
| "\\\\"                   { Buffer.add_char buffer '\\'; string_literal lexbuf }
| "\\\""                   { Buffer.add_char buffer '\"'; string_literal lexbuf }
| "\\\'"                   { Buffer.add_char buffer '\''; string_literal lexbuf }
| "\\n"                    { Buffer.add_char buffer '\n'; string_literal lexbuf }
| "\\r"                    { Buffer.add_char buffer '\r'; string_literal lexbuf }
| "\\t"                    { Buffer.add_char buffer '\t'; string_literal lexbuf } 
| "\\b"                    { Buffer.add_char buffer '\b'; string_literal lexbuf }
| "\\ "                    { Buffer.add_char buffer '\ '; string_literal lexbuf }
| "\\" ['0'-'9'] ['0'-'9'] ['0'-'9']
                           { Buffer.add_char buffer (char_for_decimal_code (Lexing.lexeme lexbuf));
        										 string_literal lexbuf }
| string as lxm            { Buffer.add_char buffer lxm; string_literal lexbuf }

(* ------- *)
(* comment *)
and comment = parse
| newline                         { token lexbuf }
| _                               { comment lexbuf }
